#include "includes.h"
#include "Time.h"



ListNode *ListCreate()
{
ListNode *Node;

Node=(ListNode *)calloc(1,sizeof(ListNode));
Node->Head=Node;
Node->Prev=Node;
Node->Next=NULL;

return(Node);
}


ListNode *MapCreate(int Buckets)
{
ListNode *Node;

Node=(ListNode *)calloc(1,sizeof(ListNode));
Node->Head=Node;
Node->Prev=Node;
Node->Next=NULL;
Node->Flags |= LIST_FLAG_HASH;
Node->Hits=Buckets;
Node->Item=calloc(Buckets, sizeof(ListNode));

return(Node);
}


void MapClear(ListNode *Map, LIST_ITEM_DESTROY_FUNC ItemDestroyer)
{
ListNode *Node;
int i;

if (Map->Flags & LIST_FLAG_HASH)
{
	Node=(ListNode *) Map->Item;
	for (i=0; i < Map->Hits; i++)
	{
		ListClear(Node, ItemDestroyer);
		Node++;
	}	
}

}


ListNode *MapGetChain(ListNode *Map, const char *Key)
{
unsigned int i;
ListNode *Node;

if (! (Map->Flags & LIST_FLAG_HASH))
{
	i=fnv_hash(Key, Map->Hits);
	Node=(ListNode *) Map->Item;
	return(Node + i);
}
return(NULL);
}



ListNode *ListGetHead(ListNode *Node)
{
if (! Node) return(NULL);

return(Node->Head);
}



unsigned long ListSize(ListNode *List)
{
ListNode *Head;
int *intptr;

if (! List) return(0);

//avoid a function call by not calling ListGetHead
Head=List->Head;
return(Head->Hits);
}


void ListSetFlags(ListNode *List, int Flags)
{
ListNode *Head;

Head=ListGetHead(List);
Head->Flags=Flags;
}

unsigned long ListSetNoOfItems(ListNode *LastItem, unsigned long val)
{
ListNode *Head;

Head=ListGetHead(LastItem);
if (LastItem->Next==NULL) 
   Head->Prev=LastItem; /* The head Item has its Prev as being the last item! */

Head->Hits=val;

return(Head->Hits);
}



unsigned long ListIncrNoOfItems(ListNode *List)
{
ListNode *Head;

//avoid a function call by not calling ListGetHead
Head=List->Head;

if (List->Next==NULL) Head->Prev=List; /* The head Item has its Prev as being the last item! */
Head->Hits++;

return(Head->Hits);
}

unsigned long ListDecrNoOfItems(ListNode *List)
{
ListNode *Head;

//avoid a function call by not calling ListGetHead
Head=List->Head;

if (List->Next==NULL) Head->Prev=List->Prev; /* The head Item has its Prev as being the last item! */
if (Head->Hits > 0) Head->Hits--;
return(Head->Hits);
}


void ListThreadNode(ListNode *Prev, ListNode *Node)
{
ListNode *Head, *Next;

//Never thread something to itself!
if (Prev==Node) return;

Next=Prev->Next;
Node->Prev=Prev;
Prev->Next=Node;
Node->Next=Next;

//avoid a function call by not calling ListGetHead
Head=Prev->Head;
Node->Head=Head;

// Next might be NULL! If it is, then our new node is last
// item in list, so update Head node accordingly
if (Next) Next->Prev=Node; 
else Head->Prev=Node;
}


void ListUnThreadNode(ListNode *Node)
{
ListNode *Head, *Prev, *Next;

//avoid a function call by not calling ListGetHead
Head=Node->Head;

//prev node of head points to LAST item in list
if (Head->Prev==Node) 
{
	Head->Prev=Node->Prev;
	if (Head->Prev==Head) Head->Prev=NULL;
}

Prev=Node->Prev;
Next=Node->Next;
if (Prev !=NULL) Prev->Next=Next;
if (Next !=NULL) Next->Prev=Prev;
if (Head->Next==Node) Head->Next=Next;
if (Head->Prev==Node) Head->Prev=Prev;
ListDecrNoOfItems(Node);
}

void ListClear(ListNode *ListStart, LIST_ITEM_DESTROY_FUNC ItemDestroyer)
{
  ListNode *Curr,*Next;

  if (! ListStart) return; 
	if (ListStart->Flags & LIST_FLAG_HASH) MapClear(ListStart, ItemDestroyer);

  Curr=ListGetNext(ListStart);
  while (Curr !=NULL) 
  {
      Next=Curr->Next;
      if (ItemDestroyer && Curr->Item) ItemDestroyer(Curr->Item);
      DestroyString(Curr->Tag);
     free(Curr);
     Curr=Next;
  }
ListStart->Next=NULL;
ListStart->Head=ListStart;
ListStart->Prev=ListStart;
ListSetNoOfItems(ListStart,0);
}


void ListDestroy(ListNode *ListStart, LIST_ITEM_DESTROY_FUNC ItemDestroyer)
{
  if (! ListStart) return; 
  ListClear(ListStart, ItemDestroyer);
  if (ListStart->Item) free(ListStart->Item);
  free(ListStart);
}



void ListAppendItems(ListNode *Dest, ListNode *Src, LIST_ITEM_CLONE_FUNC ItemCloner)
{
	ListNode *Curr;
  void *Item;

  Curr=ListGetNext(Src);
  while (Curr !=NULL) 
  {
     if (ItemCloner)
     {
	     Item=ItemCloner(Curr->Item);
	     ListAddNamedItem(Dest,Curr->Tag,Item);
     }
		else ListAddNamedItem(Dest, Curr->Tag, Curr->Item);
    Curr=ListGetNext(Curr);
  }
}


ListNode *ListClone(ListNode *ListStart, LIST_ITEM_CLONE_FUNC ItemCloner)
{
  ListNode *NewList;

  NewList=ListCreate();

	ListAppendItems(NewList, ListStart, ItemCloner);
  return(NewList);
}

ListNode *ListAddNamedItemAfter(ListNode *ListStart,const char *Name,void *Item)
{
ListNode *Curr;

if (ListStart==NULL) return(NULL);
if (ListStart->Flags & LIST_FLAG_HASH) ListStart=MapGetChain(ListStart, Name);


Curr=ListStart;
Curr->Next=(ListNode *) calloc(1,sizeof(ListNode)); 
Curr->Next->Prev=Curr;
Curr=Curr->Next;
Curr->Item=Item;
Curr->Head=ListGetHead(ListStart);
Curr->Next=NULL;
if (Name) Curr->Tag=CopyStr(NULL,Name);
Curr->Time=GetTime(TIME_CACHED);

ListIncrNoOfItems(Curr);
return(Curr);
}



ListNode *ListAddNamedItem(ListNode *ListStart,const char *Name, void *Item)
{
ListNode *Curr;

if (ListStart->Flags & LIST_FLAG_HASH) ListStart=MapGetChain(ListStart, Name);

Curr=ListGetLast(ListStart);
if (Curr==NULL) return(Curr);
return(ListAddNamedItemAfter(Curr,Name,Item));
}



ListNode *ListAddItem(ListNode *ListStart,void *Item)
{
return(ListAddNamedItem(ListStart,NULL,Item));
}




ListNode *ListInsertNamedItem(ListNode *InsertNode, const char *Name, void *Item)
{
ListNode *NewItem, *Next;

if (InsertNode->Flags & LIST_FLAG_HASH) InsertNode=MapGetChain(InsertNode, Name);
Next=InsertNode->Next;
NewItem=(ListNode *) calloc(1,sizeof(ListNode)); 
NewItem->Item=Item;
ListThreadNode(InsertNode, NewItem);
ListIncrNoOfItems(NewItem);
if (StrLen(Name)) NewItem->Tag=CopyStr(NewItem->Tag,Name);
NewItem->Time=GetTime(TIME_CACHED);

return(NewItem);
}


void OrderedListAddJump(ListNode *From, ListNode *To)
{
int result;
static int depth=0;

if (! From) return;
if (! To) return;

depth++;
if (depth > 100) 
{
	printf("ERROR: OLJ DEEEP\n");
	exit(1);
}

if (From->Jump)
{
	result=strcmp(From->Jump->Tag, To->Tag);
//	if (result > 0)  OrderedListAddJump(From->Next,To);
//	else
	if (result < 0)
	{
	OrderedListAddJump(From->Next,From->Jump);
	From->Jump=To;
	}
}
else From->Jump=To;

depth--;
}



ListNode *ListFindNamedItemInsert(ListNode *Head, const char *Name)
{
ListNode *Prev, *Curr, *Start=NULL;
int result=0, count=0;
int hops=0, jumps=0, miss=0;
unsigned int val;

if (! Head) return(Head);
if (! StrLen(Name)) return(Head);

if (Head->Flags & LIST_FLAG_HASH) Head=MapGetChain(Head, Name);

Curr=ListGetNext(Head);
if (! Curr) return(Head);

//Check last item in list
Prev=Head->Prev;
if (Prev && (Prev != Head) && Prev->Tag)
{
	if (Head->Flags & LIST_FLAG_CASE) result=strcmp(Prev->Tag,Name);
	else result=strcasecmp(Prev->Tag,Name);

	if ((Head->Flags & LIST_FLAG_ORDERED) && (result < 1)) return(Prev);
}

/*
if ((Head->Flags & LIST_FLAG_CACHE) && (Head->Jump))
{
	Prev=Head->Jump;
	if (Prev->Tag)
	{
	if (Head->Flags & LIST_FLAG_CASE) result=strcmp(Prev->Tag,Name);
	else result=strcasecmp(Prev->Tag,Name);
	if (result < 1) Curr=Prev;
	}
}
*/


Prev=Head;
while (Curr)
{
	if (Curr->Jump && Curr->Jump->Tag)
	{
		if (Head->Flags & LIST_FLAG_CASE) result=strcmp(Curr->Jump->Tag,Name);
		else result=strcasecmp(Curr->Jump->Tag,Name);
		if (result < 1) 
		{
			if ((result < 0) && Curr->Jump->Next) Curr->Jump=Curr->Jump->Next;
			Curr=Curr->Jump;
			jumps++;
 			count=0;
		}
		else miss++;
	} else if (! Start) Start=Curr;

	if (Curr->Tag)
	{
  	if (Head->Flags & LIST_FLAG_CASE) result=strcmp(Curr->Tag,Name);
		else result=strcasecmp(Curr->Tag,Name);

		if (result==0)
		{
			if (Head->Flags & LIST_FLAG_SELFORG) ListSwapItems(Curr->Prev, Curr);
			if (Head->Flags & LIST_FLAG_CACHE) Head->Jump=Curr;
			return(Curr);
		}
		if ((result > 0) && (Head->Flags & LIST_FLAG_ORDERED)) 
		{
			return(Prev);
		}
	}

hops++;
count++;


	Prev=Curr;
  Curr=ListGetNext(Curr);
}

return(Prev);
}



ListNode *ListFindNamedItem(ListNode *Head, const char *Name)
{
	ListNode *Node;
	int result;

	Node=ListFindNamedItemInsert(Head, Name);
	if ((! Node) || (Node==Head) || (! Node->Tag)) return(NULL);
 	if (Head->Flags & LIST_FLAG_CASE) result=strcmp(Node->Tag,Name);
	else result=strcasecmp(Node->Tag,Name);
	
	if (result==0) 
	{
		Node->Hits++;
		return(Node);
	}
	return(NULL);
}


ListNode *OrderedListAddNamedItem(ListNode *Head, const char *Name, void *Item)
{
ListNode *NewItem, *Prev;

Prev=ListFindNamedItemInsert(Head, Name);
NewItem=(ListNode *) calloc(1,sizeof(ListNode)); 
NewItem->Item=Item;
ListThreadNode(Prev, NewItem);
ListIncrNoOfItems(NewItem);
if (StrLen(Name)) NewItem->Tag=CopyStr(NewItem->Tag,Name);
NewItem->Time=GetTime(TIME_CACHED);

return(NewItem);
}



ListNode *ListInsertItem(ListNode *InsertNode, void *Item)
{
return(ListInsertNamedItem(InsertNode, NULL, Item));
}


ListNode *InsertItemIntoSortedList(ListNode *List, void *Item, int (*LessThanFunc)(void *, void *, void *))
{
ListNode *Curr, *Prev;

Prev=List;
Curr=ListGetNext(Prev);
while (Curr && (LessThanFunc(NULL, Curr->Item,Item)) )
{
Prev=Curr;
Curr=ListGetNext(Prev);
}

return(ListInsertItem(Prev,Item));
}


ListNode *ListGetNext(ListNode *CurrItem)
{
if (CurrItem !=NULL) return(CurrItem->Next);
else return(NULL);
}

ListNode *ListGetPrev(ListNode *CurrItem)
{
ListNode *Prev;
if (CurrItem !=NULL)
{
Prev=CurrItem->Prev;
/* Don't return the dummy header! */
if (Prev && (Prev->Prev !=NULL) && (Prev != Prev->Head)) return(Prev);
else return(NULL);
}
else return(NULL);
}


ListNode *ListGetLast(ListNode *CurrItem)
{
ListNode *Head;


Head=ListGetHead(CurrItem);
if (! Head) return(CurrItem);
/* the dummy header has a 'Prev' entry that points to the last item! */
return(Head->Prev);
}



ListNode *ListGetNth(ListNode *Head, int n)
{
ListNode *Curr;
int count=0;

if (! Head) return(NULL);

Curr=ListGetNext(Head);
while (Curr && (count < n))
{
   count++;
   Curr=ListGetNext(Curr);
}
if (count < n) return(NULL);
return(Curr);
}





ListNode *ListJoin(ListNode *List1, ListNode *List2)
{
ListNode *Curr, *StartOfList2;

Curr=List1;
/*Lists all have a dummy header!*/
StartOfList2=List2->Next;

while (Curr->Next !=NULL) Curr=Curr->Next;
Curr->Next=StartOfList2;
StartOfList2->Prev=Curr;

while (Curr->Next !=NULL) Curr=Curr->Next;
return(Curr);
}


//Item1 is before Item2!
void ListSwapItems(ListNode *Item1, ListNode *Item2)
{
ListNode *Head, *Prev, *Next;

if (! Item1) return;
if (! Item2) return;

Head=ListGetHead(Item1);
if (Head==Item1) return;
if (Head==Item2) return;

Prev=Item1->Prev;
Next=Item2->Next;

if (Head->Next==Item1) Head->Next=Item2;
if (Head->Prev==Item1) Head->Prev=Item2;

if (Prev) Prev->Next=Item2;
Item1->Prev=Item2;
Item1->Next=Next;

if (Next) Next->Prev=Item1;
Item2->Prev=Prev;
Item2->Next=Item1;

}


void ListSort(ListNode *List, void *Data, int (*LessThanFunc)(void *, void *, void *))
{
ListNode *Curr=NULL, *Prev=NULL;
int sorted=0;

while (! sorted)
{ 
  sorted=1;
  Prev=NULL;
  Curr=ListGetNext(List);
  while (Curr)
  {
    if (Prev !=NULL)
    {
       if ( (*LessThanFunc)(Data,Curr->Item,Prev->Item) )
       {
         sorted=0;
         ListSwapItems(Prev,Curr);
       }
    }
		Prev=Curr;
    Curr=ListGetNext(Curr);
  }
}

}

void ListSortNamedItems(ListNode *List)
{
ListNode *Curr=NULL, *Prev=NULL;
int sorted=0;

while (! sorted)
{ 
  sorted=1;
  Prev=NULL;
  Curr=ListGetNext(List);
  while (Curr)
  {
    if (Prev !=NULL)
    {
       if (strcmp(Prev->Tag,Curr->Tag) < 0)
       {
         sorted=0;
         ListSwapItems(Prev,Curr);
       }
    }

    Prev=Curr;
    Curr=ListGetNext(Curr);
  }
}

}





ListNode *ListFindItem(ListNode *Head, void *Item)
{
ListNode *Curr;

if (! Item) return(NULL);
Curr=ListGetNext(Head);
while (Curr)
{
	if (Curr->Item==Item) 
	{
		if (Head->Flags & LIST_FLAG_SELFORG) ListSwapItems(Curr->Prev, Curr);
		return(Curr);
	}
	Curr=ListGetNext(Curr);
}
return(Curr);
}


void *ListDeleteNode(ListNode *Node)
{
void *Contents;

if (Node==NULL) return(NULL);

ListUnThreadNode(Node);
Contents=Node->Item;

DestroyString(Node->Tag);
free(Node);
return(Contents);
}


void *ListDeleteItem(ListNode *Head, void *Item)
{
ListNode *Node;

Node=ListFindItem(Head, Item);
if (Node) ListDeleteNode(Node);

return(NULL);
}


